package it.unimi.dsi.law.triangrw;

import it.unimi.dsi.Util;
import it.unimi.dsi.bits.Fast;
import it.unimi.dsi.fastutil.booleans.BooleanArrays;
import it.unimi.dsi.fastutil.doubles.DoubleArrays;
import it.unimi.dsi.fastutil.doubles.DoubleIterators;
import it.unimi.dsi.fastutil.ints.AbstractInt2DoubleMap;
import it.unimi.dsi.fastutil.ints.Int2DoubleMap;
import it.unimi.dsi.fastutil.ints.IntArrayList;
import it.unimi.dsi.fastutil.ints.IntArrays;
import it.unimi.dsi.fastutil.io.BinIO;
import it.unimi.dsi.fastutil.io.FastBufferedOutputStream;
import it.unimi.dsi.fastutil.objects.AbstractObjectIterator;
import it.unimi.dsi.logging.ProgressLogger;
import it.unimi.dsi.util.XorShiftStarRandom;
import it.unimi.dsi.webgraph.BVGraph;
import it.unimi.dsi.webgraph.ImmutableGraph;
import it.unimi.dsi.webgraph.NodeIterator;

import java.io.DataOutputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.lang.Thread.UncaughtExceptionHandler;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicIntegerArray;

import org.apache.log4j.Logger;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPException;
import com.martiansoftware.jsap.JSAPResult;
import com.martiansoftware.jsap.Parameter;
import com.martiansoftware.jsap.SimpleJSAP;
import com.martiansoftware.jsap.Switch;
import com.martiansoftware.jsap.UnflaggedOption;

/** This class implements the Arc Label Propagation clustering algorithm as described in the paper: 
 *  <em>Arc-community detection via triangular random walks</em> by Paolo Boldi and Marco Rosa. 
 *  Given a graph, it will cluster its arcs using (implicitly, without having to build it out
 *  for true) a weighted version of its line graph; weights are established as explained in the
 *  above paper: those weights come in two flavours (ratio and mass, as discussed in the paper) and
 *  are of two types (called <var>w</var><sub>T</var> and <var>v</var><sub>T</sub>): both depend on
 *  a parameter &beta;, but while the first can be computed directly, the second requires a previous
 *  run of the {@link LinePageRank} class. 
 * 
 */
public class ALP {
	private final static Logger LOGGER = Util.getLogger( ALP.class );

	/** Maximum number of update cycles in the LP algorithm. */
	private static final int MAX_UPDATES = 20;
	
	/** The fraction of label-changing arcs under which the iteration is stopped. */
	private static final double MODIFIED_THRESHOLD = 0.001;


	/** The graph to be clustered. */
	private final ImmutableGraph graph;
	/** The transpose of the graph to be clustered. */
	private final ImmutableGraph tgraph;

	/** The number of nodes of the graph. */
	private final int n;
	/** The number of arcs of the graph. */
	private final int m;

	/** An array with one label per arc (in sequential graph-scan order). */
	private AtomicIntegerArray label;

	/** The value of the &beta; parameter (in the range [0,1]). */
	private final double beta;
	/** The value of the &alpha; parameter (in the range [0,1]). Only meaningful if we are using the <var>v</var><sub>T</sub> scheme). */
	private final double alpha;

	/** The update list for the current iteration. */
	private final int[] updateList;

	/** The array of outdegrees. */
	private final int[] outdegree;
	/** The array of indegrees. */
	private final int[] indegree;
	/** The cumulative degree: <code>offsets[i]</code> is the sum of the outdegrees of all nodes up to <code>i</code> (exclusive). */
	private final int[] offsets;


	/** The seed used to initialize the random generator. */
	private final long seed;
	/** Object used as source of randomness. */
	private final XorShiftStarRandom r;

	/** Number of threads. */
	private final int numberOfThreads;
	/** Monitor to guarantee mutual exclusion of threads. */
	private final AtomicInteger nextNode;
	/** Exception handler used by the treads. */
	private final SimpleUncaughtExceptionHandler simpleUncaughtExceptionHandler;
	/** Exception object that the threads (re)use. */
	private volatile Throwable threadException;

	/** The number of modified labels in the last run. */
	private final AtomicInteger modified;


	/** The number of the update cycle currently being executed. */
	private int update;

	/** Whether a given arc can change its label. */
	private final boolean[] canChange;

	/** The score of each arc (in case we are using the <var>v</var><sub>T</sub> scheme); it represents the PageRank
	 *  value of that arc. Otherwise, it is <code>null</code>. */
	private final double[] score;

	/** If true, we are using the ratio-triangular scheme; otherwise we are using the mass-triangular scheme. */
	private final boolean ratio;

	/** An home-made vanilla lightweight implementation of a hash-table double-valued counter. Keys in this table
	 *  are integers, and an {{@link #incr(int, double)} method is provided to increase the value associated to a given
	 *  key by a desired quantity. Objects of this class are used to make the majority computation of the labels
	 *  around each given arc.
	 * 
	 */
	private final static class OpenHashTableCounter {
		private int[] key;

		private double[] count;

		private int[] location;

		private int mask;

		private int n;

		public OpenHashTableCounter() {
			mask = -1;
			count = DoubleArrays.EMPTY_ARRAY;
			key = IntArrays.EMPTY_ARRAY;
			location = IntArrays.EMPTY_ARRAY;
		}

		public void incr( final int k, final double val ) {
			int pos = ( k * 2056437379 ) & mask;
			while ( count[ pos ] != 0 && key[ pos ] != k )
				pos = ( pos + 1 ) & mask;
			if ( count[ pos ] == 0 ) {
				key[ pos ] = k;
				location[ n++ ] = pos;
			}
			count[ pos ] += val;
		}

		private final static class Entry extends AbstractInt2DoubleMap.BasicEntry {
			public Entry() {
				super( 0, 0 );
			}

			public void setKey( final int key ) {
				this.key = key;
			}

			public double setValue( final double value ) {
				this.value = value;
				return -1; // Violates the interface, but it's all internal.
			}
		}

		public Iterator<Int2DoubleMap.Entry> entries() {
			return new AbstractObjectIterator<Int2DoubleMap.Entry>() {
				private int i;

				private Entry entry = new Entry();

				@Override
				public boolean hasNext() {
					return i < n;
				}

				@Override
				public it.unimi.dsi.fastutil.ints.Int2DoubleMap.Entry next() {
					if ( !hasNext() )
						throw new NoSuchElementException();
					final int l = location[ i++ ];
					entry.setKey( key[ l ] );
					entry.setValue( count[ l ] );
					return entry;
				}
			};
		}

		public void clear( final int size ) {
			if ( mask + 1 < ( 1 << ( Fast.ceilLog2( size ) + 1 ) ) ) {
				mask = ( 1 << ( Fast.ceilLog2( size ) + 1 ) ) - 1;
				count = new double[ mask + 1 ];
				key = new int[ mask + 1 ];
				location = new int[ mask + 1 ];
			}
			else
				while ( n-- != 0 )
					count[ location[ n ] ] = 0;
			n = 0;
		}
	}

	private final class SimpleUncaughtExceptionHandler implements UncaughtExceptionHandler {
		@Override
		public void uncaughtException( Thread t, Throwable e ) {
			threadException = e;
		}
	}

	/** Creates a new instance of the clustering algorithm. After this, one should call the {@link #computeClustering()}
	 *  to obtain the labelling.
	 * 
	 * @param graph the graph whose arcs must be clustered.
	 * @param tgraph the transpose of the graph.
	 * @param ratio if true, ratio-triangular weights must be used; otherwise, mass-triangular weights must be used. 
	 * @param alpha the &alpha; parameter, in the range [0,1] (only meaningful if <code>score</code> is provided).
	 * @param beta the &beta; parameter, in the range [0,1].
	 * @param score if not null, it contains one score for each arc that is used to compute the weights according to the <var>v</var><sub>T</sub> scheme;
	 *    if null, the <var>w</var><sub>T</sub> scheme should be used instead.
	 * @param seed the seed to be used to initialize the random generator.
	 */
	public ALP( final ImmutableGraph graph, final ImmutableGraph tgraph, final boolean ratio, final double alpha, final double beta, final double[] score, final long seed ) {
		this.graph = graph;
		this.tgraph = tgraph;
		this.n = graph.numNodes();
		this.m = (int)graph.numArcs();
		this.seed = seed;
		this.label = new AtomicIntegerArray( this.m );

		this.updateList = Util.identity( n );
		simpleUncaughtExceptionHandler = new SimpleUncaughtExceptionHandler();

		this.numberOfThreads = Runtime.getRuntime().availableProcessors();
		this.nextNode = new AtomicInteger();
		this.modified = new AtomicInteger( 0 );

		this.offsets = new int[ n ];
		this.outdegree = new int[ n ];
		LOGGER.info( "Computing outdegrees..." );
		ProgressLogger pl = new ProgressLogger( LOGGER );
		pl.expectedUpdates = n;
		pl.itemsName = "nodes";
		pl.start();
		int incr = 0;
		for ( NodeIterator iterator = graph.nodeIterator(); iterator.hasNext(); pl.lightUpdate() ) {
			final int node = iterator.nextInt();
			final int deg = iterator.outdegree();
			offsets[ node ] = incr;
			outdegree[ node ] = deg;
			incr += iterator.outdegree();
		}
		pl.done();


		this.indegree = new int[ n ];
		LOGGER.info( "Computing inndegrees..." );
		pl = new ProgressLogger( LOGGER );
		pl.expectedUpdates = n;
		pl.itemsName = "nodes";
		pl.start();
		for ( NodeIterator iterator = tgraph.nodeIterator(); iterator.hasNext(); pl.lightUpdate() ) {
			final int node = iterator.nextInt();
			final int deg = iterator.outdegree();
			indegree[ node ] = deg;
		}
		pl.done();

		this.beta = beta;

		this.r = new XorShiftStarRandom( seed );
		this.canChange = new boolean[ m ];
		this.score = score != null ? score : new double[ m ];
		if ( score == null )
			DoubleArrays.fill( this.score, 1 );
		this.alpha = alpha;
		this.ratio = ratio;
	}


	/** The version of the iteration thread used for the ratio-triangular case.
	 */
	protected final class IterationThreadRatio extends Thread {

		private final ImmutableGraph graph;

		private final ImmutableGraph tgraph;

		private final ProgressLogger pl;

		private IterationThreadRatio( final ImmutableGraph graph, final ImmutableGraph tgraph, final ProgressLogger pl ) {
			this.graph = graph;
			this.tgraph = tgraph;
			this.pl = pl;
		}

		public void run() {
			final XorShiftStarRandom r = new XorShiftStarRandom( ALP.this.seed );
			final AtomicIntegerArray label = ALP.this.label;
			final ImmutableGraph graph = this.graph;
			final ImmutableGraph tgraph = this.tgraph;
			final int n = ALP.this.n;
			final int[] updateList = ALP.this.updateList;
			final int[] outdegree = ALP.this.outdegree;
			final int[] indegree = ALP.this.indegree;
			final int granularity = Math.max( 1024, n >>> 10 );
			final double beta = ALP.this.beta;
			final int[] offsets = ALP.this.offsets;
			final double[] score = ALP.this.score;
			final int m = ALP.this.m;
			final double alpha = ALP.this.alpha;
			final boolean[] canChange = ALP.this.canChange;
			

			int start, end;
			OpenHashTableCounter map = new OpenHashTableCounter();

			for ( ;; ) {
				final int next = nextNode.getAndAdd( granularity );
				if ( next >= n ) {
					nextNode.getAndAdd( -granularity ); // Put the candle back
					return;
				}
				start = next;
				end = (int)( Math.min( n, (long)start + granularity ) );

				int sum = 0;
				for ( int k = start; k < end; k++ ) {
					final int x = updateList[ k ];
					final int degX = outdegree[ x ];
					final int innDegX = indegree[ x ];

					int[] succX = null;
					int[] innX = null;
					int[] innLabel = null;
					int[] innArcsID = null;
					boolean[][] innTriangle = null;

					sum += degX;
					for ( int i = 0; i < degX; i++ ) {
						final int currentArc = offsets[ x ] + i;
						if ( canChange[ currentArc ] ) {
							if ( succX == null ) {
								succX = graph.successorArray( x );
								innX = tgraph.successorArray( x );
								innLabel = new int[ innDegX ];
								innArcsID = new int[ innDegX ];
								innTriangle = new boolean[ innDegX ][ degX ];

								for ( int j = 0; j < innDegX; j++ ) {
									final int z = innX[ j ];
									final int degZ = outdegree[ z ];
									final int[] succZ = graph.successorArray( z );

									for ( int a = 0; a < degZ; a++ ) {
										if ( succZ[ a ] == x ) {
											innLabel[ j ] = label.get( offsets[ z ] + a );
											innArcsID[ j ] = offsets[ z ] + a;
										}
										final int idx = IntArrays.binarySearch( succX, 0, degX, succZ[ a ] );
										if ( idx >= 0 )
											innTriangle[ j ][ idx ] = true;
									}
								}
							}

							final int y = succX[ i ];
							final int degY = outdegree[ y ];
							final int[] succY = graph.successorArray( y );
							

							final int currentLabel = label.get( currentArc );
							map.clear( degY + innDegX );
							int a, b;
							// /// inserting in maps
							for ( a = 0, b = 0; a < degX && b < degY; ) {
								if ( succX[ a ] < succY[ b ] )
									a++;
								else if ( succY[ b ] < succX[ a ] ) {
									map.incr( label.get( offsets[ y ] + b ), score[ currentArc ] * ( ( alpha * beta ) + ( 1 - alpha ) / m )  );
									b++;
								}
								else {
									map.incr( label.get( offsets[ y ] + b ), score[ currentArc ] * ( ( alpha * 1 ) + ( 1 - alpha ) / m ) );
									a++;
									b++;
								}
							}
							for ( ; b < degY; b++ )
								map.incr( label.get( offsets[ y ] + b ), score[ currentArc ] * ( ( alpha * beta ) + ( 1 - alpha ) / m ) );


							// /// inserting in maps
							for ( a = 0; a < innDegX; a++ ) {
								final int l = innLabel[ a ];
								if ( innTriangle[ a ][ i ] )
									map.incr( l, 1 );
								else
									map.incr( l, beta );
							}


							double max = Double.NEGATIVE_INFINITY;
							IntArrayList majorities = new IntArrayList();

							for ( Iterator<Int2DoubleMap.Entry> entries = map.entries(); entries.hasNext(); ) {
								Int2DoubleMap.Entry entry = entries.next();
								final int l = entry.getIntKey();
								final double val = entry.getDoubleValue();

								if ( max == val )
									majorities.add( l );

								if ( max < val ) {
									majorities.clear();
									max = val;
									majorities.add( l );
								}
							}


							if ( majorities.isEmpty() )
								majorities.add( currentLabel );

							if ( majorities.size() < 2 )
								canChange[ currentArc ] = false;

							final int nextLabel = majorities.getInt( r.nextInt( majorities.size() ) );
							if ( nextLabel != currentLabel ) {
								modified.addAndGet( 1 );
								label.set( currentArc, nextLabel );
								for ( a = 0; a < innDegX; a++ )
									canChange[ innArcsID[ a ] ] = true;
								for ( b = 0; b < degY; b++ )
									canChange[ offsets[ y ] + b ] = true;
							}
						}
					}
				}
				synchronized ( pl ) {
					pl.update( sum );
				}
			}
		}
	}


	/** The version of the iteration thread used for the mass-triangular case.
	 */
	protected final class IterationThreadMass extends Thread {

		private final ImmutableGraph graph;

		private final ImmutableGraph tgraph;

		private final ProgressLogger pl;

		private IterationThreadMass( final ImmutableGraph graph, final ImmutableGraph tgraph, final ProgressLogger pl ) {
			this.graph = graph;
			this.tgraph = tgraph;
			this.pl = pl;
		}

		public void run() {
			final XorShiftStarRandom r = new XorShiftStarRandom( ALP.this.seed );
			final AtomicIntegerArray label = ALP.this.label;
			final ImmutableGraph graph = this.graph;
			final ImmutableGraph tgraph = this.tgraph;
			final int n = ALP.this.n;
			final int[] updateList = ALP.this.updateList;
			final int[] outdegree = ALP.this.outdegree;
			final int[] indegree = ALP.this.indegree;
			final int granularity = Math.max( 1024, n >>> 10 );
			final double beta = ALP.this.beta;
			final double[] score = ALP.this.score;
			final int[] offsets = ALP.this.offsets;
			final int m = ALP.this.m;
			final double alpha = ALP.this.alpha;
			final boolean[] canChange = ALP.this.canChange;
			int start, end;
			OpenHashTableCounter map = new OpenHashTableCounter();

			for ( ;; ) {
				final int next = nextNode.getAndAdd( granularity );
				if ( next >= n ) {
					nextNode.getAndAdd( -granularity ); // Put the candle back
					return;
				}
				start = next;
				end = (int)( Math.min( n, (long)start + granularity ) );

				int sum = 0;
				for ( int k = start; k < end; k++ ) {
					final int x = updateList[ k ];
					final int degX = outdegree[ x ];
					final int innDegX = indegree[ x ];

					int[] succX = null;
					int[] innX = null;
					int[] innLabel = null;
					int[] innArcsID = null;
					int[] innTr = null;
					int[] innNonTr = null;
					boolean[][] innTriangle = null;

					sum += degX;
					for ( int i = 0; i < degX; i++ ) {
						final int currentArc = offsets[ x ] + i;
						if ( canChange[ currentArc ] ) {
							if ( succX == null ) {
								succX = graph.successorArray( x );
								innX = tgraph.successorArray( x );
								innLabel = new int[ innDegX ];
								innArcsID = new int[ innDegX ];
								innTr = new int[ innDegX ];
								innNonTr = new int[ innDegX ];
								innTriangle = new boolean[ innDegX ][ degX ];

								for ( int j = 0; j < innDegX; j++ ) {
									final int z = innX[ j ];
									final int degZ = outdegree[ z ];
									final int[] succZ = graph.successorArray( z );

									int a, b;
									// /// calculating the masses
									for ( a = 0, b = 0; a < degZ && b < degX; ) {
										if ( succZ[ a ] == x ) {
											innLabel[ j ] = label.get( offsets[ z ] + a );
											innArcsID[ j ] = offsets[ z ] + a;
										}
										final int idx = IntArrays.binarySearch( succX, 0, degX, succZ[ a ] );
										if ( idx >= 0 )
											innTriangle[ j ][ idx ] = true;

										if ( succZ[ a ] < succX[ b ] )
											a++;
										else if ( succX[ b ] < succZ[ a ] ) {
											innNonTr[ j ]++;
											b++;
										}
										else {
											innTr[ j ]++;
											a++;
											b++;
										}
									}

									for ( ; a < degZ; a++ ) {
										if ( succZ[ a ] == x ) {
											innLabel[ j ] = label.get( offsets[ z ] + a );
											innArcsID[ j ] = offsets[ z ] + a;
										}
										final int idx = IntArrays.binarySearch( succX, 0, degX, succZ[ a ] );
										if ( idx >= 0 )
											innTriangle[ j ][ idx ] = true;
									}

									for ( ; b < degX; b++ )
										innNonTr[ j ]++;
								}
							}

							final int y = succX[ i ];
							final int degY = outdegree[ y ];
							final int[] succY = graph.successorArray( y );

							double tr = 0;
							double nonTr = 0;

							final int currentLabel = label.get( currentArc );
							map.clear( degY + innDegX );
							int a, b;
							// //// calculating the masses
							for ( a = 0, b = 0; a < degX && b < degY; ) {
								if ( succX[ a ] < succY[ b ] )
									a++;
								else if ( succY[ b ] < succX[ a ] ) {
									nonTr++;
									b++;
								}
								else {
									tr++;
									a++;
									b++;
								}
							}
							for ( ; b < degY; b++ )
								nonTr++;


							double tmpBeta = beta;
							if ( tr < 1 ) tmpBeta = 1;
							if ( nonTr < 1 ) tmpBeta = 0;
							 	
							// //// inserting in maps
							for ( a = 0, b = 0; a < degX && b < degY; ) {
								if ( succX[ a ] < succY[ b ] )
									a++;
								else if ( succY[ b ] < succX[ a ] ) {
									map.incr( label.get( offsets[ y ] + b ), score[ currentArc ] * ( ( alpha * tmpBeta / nonTr ) + ( 1 - alpha ) / m ) );
									b++;
								}
								else {
									map.incr( label.get( offsets[ y ] + b ), score[ currentArc ] * ( ( alpha * ( 1.0 - tmpBeta ) / tr ) + ( 1 - alpha ) / m ) );
									a++;
									b++;
								}
							}
							for ( ; b < degY; b++ )
								map.incr( label.get( offsets[ y ] + b ), score[ currentArc ] * ( ( alpha * tmpBeta / nonTr ) + ( 1 - alpha ) / m ) );


							// //// inserting in maps
							for ( a = 0; a < innDegX; a++ ) {
								final int l = innLabel[ a ];
								tmpBeta = beta;
								if ( innTr[ a ] < 1 ) tmpBeta = 1;
								if ( innNonTr[ a ] < 1 ) tmpBeta = 0;
									
								if ( innTriangle[ a ][ i ] )
									map.incr( l, score[ innArcsID[ a ] ] * ( ( alpha * ( 1.0 - tmpBeta ) / innTr[ a ] ) + ( 1 - alpha ) / m ) );
								else
									map.incr( l, score[ innArcsID[ a ] ] * ( ( alpha * tmpBeta / innNonTr[ a ] ) + ( 1 - alpha ) / m ) );
							}


							double max = Double.NEGATIVE_INFINITY;
							IntArrayList majorities = new IntArrayList();

							for ( Iterator<Int2DoubleMap.Entry> entries = map.entries(); entries.hasNext(); ) {
								Int2DoubleMap.Entry entry = entries.next();
								final int l = entry.getIntKey();
								final double val = entry.getDoubleValue();

								if ( max == val )
									majorities.add( l );

								if ( max < val ) {
									majorities.clear();
									max = val;
									majorities.add( l );
								}
							}


							if ( majorities.isEmpty() )
								majorities.add( currentLabel );

							if ( majorities.size() < 2 )
								canChange[ currentArc ] = false;

							final int nextLabel = majorities.getInt( r.nextInt( majorities.size() ) );
							if ( nextLabel != currentLabel ) {
								modified.addAndGet( 1 );
								label.set( currentArc, nextLabel );
								for ( a = 0; a < innDegX; a++ )
									canChange[ innArcsID[ a ] ] = true;
								for ( b = 0; b < degY; b++ )
									canChange[ offsets[ y ] + b ] = true;
							}
						}
					}
				}
				synchronized ( pl ) {
					pl.update( sum );
				}
			}
		}
	}

	/** Performs a new, complete update cycle. Note that the variable {@link #update} is not increased.
	 * 
	 */
	private void update() {
		final int m = this.m;
		final int[] updateList = this.updateList;
		modified.set( 0 );

		IntArrays.shuffle( updateList, r );

		ProgressLogger pl = new ProgressLogger( LOGGER );
		pl.expectedUpdates = m;
		pl.logInterval = ProgressLogger.TEN_SECONDS;
		pl.itemsName = "arcs";
		pl.start( "Starting update " + ( update + 1 ) + "..." );


		final Thread[] thread = new Thread[ numberOfThreads ];
		nextNode.set( 0 );
		for ( int i = 0; i < numberOfThreads; i++ ) {
			thread[ i ] = ratio ? new IterationThreadRatio( graph.copy(), tgraph.copy(), pl ) : new IterationThreadMass( graph.copy(), tgraph.copy(), pl );
			thread[ i ].setUncaughtExceptionHandler( simpleUncaughtExceptionHandler );
			thread[ i ].start();
		}
		for ( int i = 0; i < numberOfThreads; i++ ) {
			try {
				thread[ i ].join();
			}
			catch ( InterruptedException e ) {
				throw new RuntimeException( e );
			}
		}
		if ( threadException != null )
			throw new RuntimeException( threadException );
		pl.done();
	}


	/** Computes the clustering, returning the array of labels (one label for each arc, in the order in which
	 *  arcs are returned by a sequential scan of the graph). This iteration is stopped when no more than {@link #MODIFIED_THRESHOLD}
	 *  fraction of arcs change label, and in any case no more than {@link #MAX_UPDATES} updates are performed. 
	 * 
	 * @return the clustering.
	 */
	public AtomicIntegerArray computeClustering() {
		final int m = this.m;

		Util.identity( updateList );
		for ( int i = 0; i < m; i++ )
			label.set( i, i );
		BooleanArrays.fill( canChange, true );

		final ProgressLogger pl = new ProgressLogger( LOGGER, ProgressLogger.TEN_SECONDS, "updates" );
		pl.start( "Starting ..." );

		update = 0;
		do {
			update();
			pl.updateAndDisplay();
			LOGGER.info( "Modified: " + modified.get() + " (" + modified.get() * 100.0 / m + "%)" );
			update++;
		} while ( modified.get() > MODIFIED_THRESHOLD * m  && update < MAX_UPDATES );


		pl.done();
		LOGGER.info( "Completed." );

		return label;
	}

	
	public static void main( String[] args ) throws IOException, JSAPException {
		SimpleJSAP jsap = new SimpleJSAP( ALP.class.getName(), "Computes Raghavan's clusters on a line graph.", new Parameter[] {
				new FlaggedOption( "randomSeed", JSAP.LONG_PARSER, "0", JSAP.NOT_REQUIRED, 'r', "random-seed", "The random seed." ),
				new FlaggedOption( "beta", JSAP.DOUBLE_PARSER, "0.01", JSAP.NOT_REQUIRED, 'b', "beta", "The beta value." ),
				new FlaggedOption( "score", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 's', "score", "A scoring function for each arc." ),
				new FlaggedOption( "alpha", JSAP.DOUBLE_PARSER, "1", JSAP.NOT_REQUIRED, 'a', "alpha", "Damping factor if score is a pagerank score." ),
				new Switch( "ratio", 't', "ratio", "If specified, a ratio triangular weighting scheme will be used." ),
				new UnflaggedOption( "graph", JSAP.STRING_PARSER, JSAP.REQUIRED, "The basename of loopless graph." ),
				new UnflaggedOption( "tgraph", JSAP.STRING_PARSER, JSAP.REQUIRED, "The basename of the transposed version of the input graph." ),
				new UnflaggedOption( "dest", JSAP.STRING_PARSER, JSAP.REQUIRED, "The output labelling." ), } );

		JSAPResult jsapResult = jsap.parse( args );
		if ( jsap.messagePrinted() )
			System.exit( 1 );

		double[] score = jsapResult.userSpecified( "score" ) ? DoubleIterators.unwrap( BinIO.asDoubleIterator( jsapResult.getString( "score" ) ) ) : null;
		ALP clustering = new ALP( BVGraph.load( jsapResult.getString( "graph" ) ), BVGraph.load( jsapResult.getString( "tgraph" ) ), 
				jsapResult.getBoolean( "ratio" ), jsapResult.getDouble( "alpha" ),  jsapResult.getDouble( "beta" ), score, jsapResult.getLong( "randomSeed" ) );

		AtomicIntegerArray label = clustering.computeClustering();
		DataOutputStream out = new DataOutputStream( new FastBufferedOutputStream( new FileOutputStream( jsapResult.getString( "dest" ) ) ) );
		for ( int i = 0; i < label.length(); i++ )
			out.writeInt( label.get( i ) );

		out.close();
	}
}
