package it.unimi.dsi.law.triangrw;

/*		 
 * Copyright (C) 2011 Paolo Boldi and Marco Rosa 
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 3 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 */

import it.unimi.dsi.Util;
import it.unimi.dsi.fastutil.ints.IntOpenHashSet;
import it.unimi.dsi.fastutil.io.BinIO;
import it.unimi.dsi.fastutil.io.FastBufferedOutputStream;
import it.unimi.dsi.io.InputBitStream;
import it.unimi.dsi.io.OutputBitStream;
import it.unimi.dsi.logging.ProgressLogger;
import it.unimi.dsi.webgraph.BVGraph;
import it.unimi.dsi.webgraph.ImmutableGraph;
import it.unimi.dsi.webgraph.LazyIntIterator;
import it.unimi.dsi.webgraph.NodeIterator;
import it.unimi.dsi.webgraph.labelling.ArcLabelledImmutableGraph;
import it.unimi.dsi.webgraph.labelling.BitStreamArcLabelledImmutableGraph;
import it.unimi.dsi.webgraph.labelling.Label;

import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;

import org.apache.log4j.Logger;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPException;
import com.martiansoftware.jsap.JSAPResult;
import com.martiansoftware.jsap.Parameter;
import com.martiansoftware.jsap.SimpleJSAP;
import com.martiansoftware.jsap.Switch;
import com.martiansoftware.jsap.UnflaggedOption;


/** The main method of this class, given a graph <var>G</var> and its line graph <var>L</var> (that must have been computed beforehand,
 *  typically using <code>it.unimi.dsi.webgraph.Transform line</code>), computes the weights to be assigned to the arcs of <var>L</var>
 *  according to a mass-triangular random walk (see the paper <em>Arc-community detection via triangular random walks</em> by
 *  Paolo Boldi and Marco Rosa). The labels will be written in the form of a {@link ArcLabelledImmutableGraph}, whose underlying graph
 *  is <var>L</var>. Optionally, one can provide an additional file containing one double per node of <var>L</var> (called the score
 *  of that node): in this case, each weight will be obtained applying the formula for the <var>v</var><sub>T</sub> scheme, as explained
 *  in the paper (the scores, in that case, represent the PageRank computed for that node of the line graph).
 */
public class BuildTriangularWeightedLineGraph {
	public static final boolean DEBUG = false;
	private static final Logger LOGGER = Util.getLogger( BuildTriangularWeightedLineGraph.class ); 
	{
		Util.ensureLog4JIsConfigured();
	}
	
	/** An abstract (single-attribute) float label. This class provides basic methods for a label holding a floating-point number.
	 * Concrete implementations may impose further requirements on the number.
	 * 
	 * <p>Implementing subclasses must provide constructors, {@link Label#copy()},
	 * {@link Label#fromBitStream(it.unimi.dsi.io.InputBitStream, int)}, {@link Label#toBitStream(OutputBitStream, int)}
	 * and possibly override {@link #toString()}.
	 */
	public static abstract class AbstractFloatLabel extends it.unimi.dsi.webgraph.labelling.AbstractLabel implements it.unimi.dsi.webgraph.labelling.Label {
		/** The key of the attribute represented by this label. */
		protected final String key;
		/** The value of the attribute represented by this label. */
		public float value;

		/** Creates an int label with given key and value.
		 * 
		 * @param key the (only) key of this label.
		 * @param value the value of this label.
		 */
		public AbstractFloatLabel( String key, float value ) {
			this.key = key;
			this.value = value;
		}
		@Override
		public boolean equals( Object x ) {
			if ( x instanceof AbstractFloatLabel ) return ( value == ( (AbstractFloatLabel)x ).value );
			else return false;
		}
		@Override
		public int hashCode() {
			return new Float(value).hashCode();
		}
		public String wellKnownAttributeKey() { return key; }
		public String[] attributeKeys() { return new String[] { key }; }
		@SuppressWarnings({ "rawtypes", "unchecked" })
		public Class[] attributeTypes() { return new Class[] { int.class }; }
		public Object get( String theKey ) { return Float.valueOf( getFloat( theKey ) ); }
		public int getInt( String theKey ) { return (int)getFloat( theKey ); }
		public long getLong( String theKey ) { return (long)getFloat( theKey ); }
		public float getFloat( String theKey ) { if ( key.equals( theKey ) ) return value; throw new IllegalArgumentException(); }
		public double getDouble( String theKey ) { return (double)getFloat( theKey ); }
		public Object get() { return Integer.valueOf( getInt() ); }
		public int getInt() { return (int)value; }
		public long getLong() { return (long)value; }
		public float getFloat() { return value; }
		public double getDouble() { return value; }
		public String toString() { return key + ":" + value; }
	}

	/** An integer represented in fixed width. The provided width must
	 * be smaller than 32.
	 */
	@SuppressWarnings("unchecked")
	public static class FixedWidthFloatLabel extends AbstractFloatLabel {
		/** Creates a new fixed-width float label.
		 * 
		 * @param key the (only) key of this label.
		 * @param value the value of this label.
		 */
		public FixedWidthFloatLabel( String key, float value ) {
			super( key, value );
		}

		/** Creates a new fixed-width int label of value 0.
		 * 
		 * @param key the (only) key of this label.
		 */
		public FixedWidthFloatLabel( String key ) {
			this( key, (float)0 );
		}
		
		public FixedWidthFloatLabel( String... arg ) {
			this( arg[ 0 ], arg.length == 2 ? Float.parseFloat( arg[ 1 ] ) : 0 );
		}

		public FixedWidthFloatLabel copy() { return new FixedWidthFloatLabel( key, value ); }

		public int fromBitStream( final InputBitStream inputBitStream, final int sourceUnused ) throws IOException {
			value = Float.intBitsToFloat( inputBitStream.readInt( Integer.SIZE ) );
			return Integer.SIZE;
		}

		public int toBitStream( final OutputBitStream outputBitStream, final int sourceUnused ) throws IOException {
			return outputBitStream.writeInt(Float.floatToIntBits(value), Integer.SIZE );
		}

		public String toString() { return key + ":" + value; }
		public String toSpec() { return this.getClass().getName() + "(" + key + ")"; }

		public int fixedWidth() {
			return Float.SIZE;
		}
	}

	public static void main( String[] args ) throws IOException, JSAPException {
		SimpleJSAP jsap = new SimpleJSAP( BuildTriangularWeightedLineGraph.class.getName(), "Given a graph and the corresponding line graph, builds " +
				"a weighted version of the latter with weights suitable for a mass-triangular random walk. Optionally, the weight of arc (x,y), say w(x,y), is modified " +
				"by a value associated to x, called its score, say s(x), according to the formula s(x)*(a*w(x,y)+(1-a)/n), where a is the damping factor.", new Parameter[] {
			new FlaggedOption( "beta", JSAP.DOUBLE_PARSER, "0.2", JSAP.NOT_REQUIRED, 'b', "beta", "The beta parameter." ),
			new Switch( "normalRW", 'n', "Use a normal random walk." ),
			new FlaggedOption( "scoreFile", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 's', "scoreFile", "A file containing one score per node." ),
			new FlaggedOption( "alpha", JSAP.DOUBLE_PARSER, JSAP.NO_DEFAULT, JSAP.NOT_REQUIRED, 'a', "alpha", "The value of alpha (to be specified along with the option -s)" ),
			new UnflaggedOption( "graph", JSAP.STRING_PARSER, JSAP.REQUIRED, "The basename of the graph." ),
			new UnflaggedOption( "lineGraph", JSAP.STRING_PARSER, JSAP.REQUIRED, "The basename of the line graph." ),
			new UnflaggedOption( "mapBasename", JSAP.STRING_PARSER, JSAP.REQUIRED, "The basename of the source and target map for the line graph." ),
			new UnflaggedOption( "resultGraph", JSAP.STRING_PARSER, JSAP.REQUIRED, "The basename of the resulting graph." )
		} );

		JSAPResult jsapResult = jsap.parse( args );
		if ( jsap.messagePrinted() ) System.exit( 1 );
		
		String basename = jsapResult.getString( "graph" );
		String lineBasename = jsapResult.getString( "lineGraph" );
		BVGraph g = BVGraph.load( basename );

		double alpha = 0;
		final double beta = jsapResult.getDouble( "beta" );
		final boolean normalRW = jsapResult.userSpecified( "normalRW" );		
		
		boolean scores = jsapResult.userSpecified( "scoreFile" );
		double score[];

		if ( scores && !jsapResult.userSpecified( "alpha" ) ) {
			System.err.println( "In the presence of a score file, you must also specify a value for alpha" );
			System.exit( 1 );
		}
		if ( scores ) {
			score = BinIO.loadDoubles( jsapResult.getString( "scoreFile" ) );
			alpha = jsapResult.getDouble( "alpha" );	
		}
		else score = null;
		

		LOGGER.info( "Computing the degree vector" );
		final int[] degree = new int[ g.numNodes() ];
		int m = 0;
		for( NodeIterator iterator = g.nodeIterator(); iterator.hasNext(); ) {
			final int node = iterator.nextInt();
			final int deg = iterator.outdegree();
			degree[ node ] = deg;
			m += deg * deg;
		}
	
		LOGGER.info( "Loading the line graph" );
		ImmutableGraph line = ImmutableGraph.load( lineBasename );
		final int n = line.numNodes();
		int[] source = BinIO.loadInts( jsapResult.getString( "mapBasename" ) + ".source" );
		int[] target = BinIO.loadInts( jsapResult.getString( "mapBasename" ) + ".target" );
	
		String resultBasename = jsapResult.getString( "resultGraph" );
		String labelBasename = resultBasename + BitStreamArcLabelledImmutableGraph.LABELS_EXTENSION;
		
		int prevSource = -1;
		IntOpenHashSet edges = new IntOpenHashSet();
		
		double valOther,valTriang;
		int i,node,deg,x,y,z,intersection;
		int[] succ;
		
		ProgressLogger pl = new ProgressLogger();
		pl.expectedUpdates = n;
		pl.start( "Computing weights" );
		
		OutputBitStream obs = new OutputBitStream( new FastBufferedOutputStream( new FileOutputStream( labelBasename ) ) );
		FixedWidthFloatLabel label = new FixedWidthFloatLabel( "value" );
		
		for( NodeIterator iterator = line.nodeIterator(); iterator.hasNext(); pl.lightUpdate() ) {
			node = iterator.nextInt();
			deg = iterator.outdegree();
			succ = iterator.successorArray();
			
			x = source[ node ];
			y = target[ node ];
			intersection = 0;
			
			if( prevSource != x) {
				prevSource = x;
				edges.clear();
				LazyIntIterator succIntIterator = g.successors( x );
				for( i = 0; i < degree[ x ]; i++ )
					edges.add( succIntIterator.nextInt() );
			}
			
							

			for( i = 0; i < deg; i++ ) {
				z = target[ succ[ i ] ];
				if ( z != x && edges.contains( z ) ) intersection++;
			}
	
			if( intersection == 0 || normalRW ) {
				valOther = 1.0 / deg;
				for( i = 0; i < deg; i++ ) { 					
					label.value = (float)valOther;
					if ( scores ) label.value *= score[ node ];
					label.toBitStream( obs, node );
				}				
			} else {
				for( i = 0; i < deg; i++ ) {
					valTriang = (1-beta)/intersection;
					valOther = beta/(degree[ y ] - intersection);
					
					if( edges.contains( target[ succ[ i ] ] ) )
						label.value = (float)valTriang;
					else
						label.value = (float)valOther;
					if ( scores ) label.value = (float) ( score[ node ] * ( alpha * label.value + ( 1 - alpha ) / n ) );
					label.toBitStream( obs, node );
				}
			}
		}		
		obs.close();
		pl.done();
		
		LOGGER.info( "Writing the property file" );
		PrintWriter pw = new PrintWriter( new FileWriter( resultBasename + ImmutableGraph.PROPERTIES_EXTENSION ) );
		pw.println( ImmutableGraph.GRAPHCLASS_PROPERTY_KEY + " = " + BitStreamArcLabelledImmutableGraph.class.getName() );
		pw.println( BitStreamArcLabelledImmutableGraph.LABELSPEC_PROPERTY_KEY + " = " + FixedWidthFloatLabel.class.getName() + "(value)" );
		pw.println( ArcLabelledImmutableGraph.UNDERLYINGGRAPH_PROPERTY_KEY + " = " + lineBasename );
		pw.close();
		
	}
}

