package it.unimi.dsi.law.triangrw;

/*		 
 * Copyright (C) 2011 Paolo Boldi and Marco Rosa 
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 3 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 */

import it.unimi.dsi.Util;
import it.unimi.dsi.fastutil.ints.IntOpenHashSet;
import it.unimi.dsi.fastutil.ints.IntSet;
import it.unimi.dsi.fastutil.io.BinIO;
import it.unimi.dsi.fastutil.longs.Long2ObjectMap;
import it.unimi.dsi.fastutil.longs.Long2ObjectOpenHashMap;
import it.unimi.dsi.fastutil.longs.Long2ObjectMap.Entry;
import it.unimi.dsi.fastutil.objects.Object2LongFunction;
import it.unimi.dsi.fastutil.objects.ObjectIterator;
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet;
import it.unimi.dsi.fastutil.objects.ObjectSet;
import it.unimi.dsi.logging.ProgressLogger;

import java.io.IOException;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.apache.log4j.Logger;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPException;
import com.martiansoftware.jsap.JSAPResult;
import com.martiansoftware.jsap.Parameter;
import com.martiansoftware.jsap.SimpleJSAP;
import com.martiansoftware.jsap.UnflaggedOption;

/** Extract titles form the DBLP file (for more information, please 
 *  use the "--help" option to the main).
 */
public class ExtractTitles extends DefaultHandler {
	private static final Logger LOGGER = Util.getLogger( ExtractTitles.class ); 
	{
		Util.ensureLog4JIsConfigured();
	}

	private StringBuilder currentTitle = new StringBuilder();
	private StringBuilder currentAuthor = new StringBuilder();
	private IntSet authors = new IntOpenHashSet();
	private Long2ObjectMap<String> coauthorMap = new Long2ObjectOpenHashMap<String>();
	{
		coauthorMap.defaultReturnValue( "" );
	}
	private String title;
	private static final ObjectSet<String> publicationElement = new ObjectOpenHashSet<String>( 
			new String[] { "article", "inproceedings", "proceedings", "book", "incollection", "phdthesis", "mastersthesis", "www" } );
	
	private boolean inTitle = false;
	private boolean inAuthor = false;
	private final Object2LongFunction<String> authorMap;
	private final ProgressLogger pl;
	
	public ExtractTitles( Object2LongFunction<String> authorMap, ProgressLogger pl ) {
		this.authorMap = authorMap;
		this.pl = pl;
	}

	@Override
	public void startElement( String uri, String localName, String qName, Attributes attributes ) {
		if ( publicationElement.contains( qName ) ) {
			pl.lightUpdate();
			authors.clear();
			title = "";
		} else if ( qName.equals( "title" ) ) {
			inTitle = true;
		} else if ( qName.equals( "author" ) ) {
			inAuthor = true;
		} 
	}
	
	@Override
	public void endElement( String uri, String localName, String qName ) {
		if ( publicationElement.contains( qName ) ) {
			//System.out.println( "Publication: " );
			//System.out.println( "\tTitle: " + title );
			//System.out.println( "\tAuthor(s): " + authors );
			for ( int x: authors )
				for ( int y: authors ) 
					if ( x < y ) {
						long v = ( (long)x << 32 ) | y;
						coauthorMap.put( v, coauthorMap.get( v ) + " " + title );
					}
		} else if ( qName.equals( "title" ) && inTitle ) {
			inTitle = false;
			title = currentTitle.toString();
			currentTitle.setLength( 0 );
		} else if ( qName.equals( "author" ) && inAuthor ) {
			inAuthor = false;
			authors.add( (int)authorMap.getLong( currentAuthor.toString() ) );
			currentAuthor.setLength( 0 );
		}
	}

	@Override
	public void characters( char[] ch, int start, int length ) throws SAXException {
		if ( inTitle ) currentTitle.append( ch, start, length );
		else if ( inAuthor ) currentAuthor.append( ch, start, length );
	}
	
	public static class CountPublications extends DefaultHandler {
		public int nPub;
		private final ProgressLogger pl;
		public CountPublications( ProgressLogger pl ) {
			this.pl = pl;
		}
		@Override
		public void startElement( String uri, String localName, String qName, Attributes attributes ) {
			if ( publicationElement.contains( qName ) ) {
				nPub++;
				pl.lightUpdate();
			}
		}
		
		
	}
	

	@SuppressWarnings( "unchecked" )
	public static void main( String arg[] ) throws ParserConfigurationException, SAXException, IOException, ClassNotFoundException, JSAPException {
		final SimpleJSAP jsap = new SimpleJSAP( ExtractTitles.class.getName(), "Parses a DBLP XML file and extracts the list of coauthored papers. "+
				"For each pair of authors that coauthored (at least) one paper, an output line of the form x\\ty\\tT will be produced, " +
				"where (x,y) are the number of the two authors (x smaller than y) and T is the concatenation of the titles of the coauthored papers.",
				new Parameter[] {
					new UnflaggedOption( "dblpXMLFile", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.REQUIRED, JSAP.NOT_GREEDY, "The XML DBLP file to be parsed." ),
					new UnflaggedOption( "authorMap", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.REQUIRED, JSAP.NOT_GREEDY, "The XML DBLP file to be parsed." )
		});

		JSAPResult jsapResult = jsap.parse( arg );
		if ( jsap.messagePrinted() ) return;
		

		String xmlFilename = jsapResult.getString( "dblpXMLFile" );

		SAXParser parser = SAXParserFactory.newInstance().newSAXParser();
		
		ProgressLogger pl = new ProgressLogger( LOGGER );
		pl.start( "Counting publications" );
		CountPublications countPublications = new CountPublications( pl );
		parser.parse( xmlFilename, countPublications );
		pl.done();
		
		Object2LongFunction<String> authorMap = (Object2LongFunction<String>)BinIO.loadObject( jsapResult.getString( "authorMap" ) );
		pl.itemsName = "publications";
		pl.expectedUpdates = countPublications.nPub;
		pl.start( "Extracting titles" );
		ExtractTitles extractTitles = new ExtractTitles( authorMap, pl );
		parser.parse( xmlFilename, extractTitles );
		pl.done();

		pl.itemsName = "author pairs";
		pl.expectedUpdates = extractTitles.coauthorMap.size();
		pl.start( "Producing output" );
		ObjectIterator<Entry<String>> iterator = extractTitles.coauthorMap.long2ObjectEntrySet().iterator();
		while ( iterator.hasNext() ) {
			Entry<String> pair = iterator.next();
			int x = (int)( pair.getLongKey() >> 32 );
			int y = (int)( pair.getLongKey() & 0xFFFFFFFF );
			System.out.println( x + "\t" + y + "\t" + pair.getValue() );
			pl.lightUpdate();
		}
		pl.done();
	}
	
}
