package it.unimi.dsi.law.triangrw;

import it.unimi.dsi.Util;
import it.unimi.dsi.fastutil.doubles.DoubleArrays;
import it.unimi.dsi.fastutil.io.BinIO;
import it.unimi.dsi.logging.ProgressLogger;
import it.unimi.dsi.webgraph.BVGraph;
import it.unimi.dsi.webgraph.ImmutableGraph;
import it.unimi.dsi.webgraph.NodeIterator;

import java.io.IOException;
import java.lang.Thread.UncaughtExceptionHandler;
import java.util.concurrent.atomic.AtomicInteger;

import org.apache.log4j.Logger;

import com.martiansoftware.jsap.FlaggedOption;
import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPException;
import com.martiansoftware.jsap.JSAPResult;
import com.martiansoftware.jsap.Parameter;
import com.martiansoftware.jsap.SimpleJSAP;
import com.martiansoftware.jsap.Switch;
import com.martiansoftware.jsap.UnflaggedOption;

/** This class computes the weighted PageRank on a line graph whose weights are those
 *  described in the paper:  <em>Arc-community detection via triangular random walks</em> by Paolo Boldi and Marco Rosa. 
 *  Given a graph, it will assign a PageRank value to each of its arcs using (implicitly, without having to build it out
 *  for true) a weighted version of its line graph; weights are established as explained in the
 *  above paper: those weights (called <var>w</var><sub>T</var> in the paper),  
 *  come in two flavours (ratio and mass, as discussed in the paper). The output of
 *  this class can then be provided as input to {@link ALP} if the latter is to be
 *  used with the <var>v</var><sub>T</sub> weighting scheme. 
 */
public class LinePageRank {
	private final static Logger LOGGER = Util.getLogger( LinePageRank.class );

	/** The maximum number of PageRank iterations. */
	private final int maxIter;

	/** The graph whose arcs must be scored. */
	private final ImmutableGraph graph;

	/** The number of nodes. */
	private final int n;
	/** The number of arcs. */
	private final int m;

	/** The array of outdegrees. */
	private final int[] outdegree;
	/** The cumulative degree: <code>cumul[i]</code> is the sum of the outdegrees of all nodes up to <code>i</code> (exclusive). */
	private final int[] cumul;

	/** Monitor to guarantee mutual exclusion of threads. */	
	private final AtomicInteger nextNode;
	/** The number of threads. */
	private final int numberOfThreads;
	/** Exception handler used by the treads. */
	private final SimpleUncaughtExceptionHandler simpleUncaughtExceptionHandler;
	/** Exception object that the threads (re)use. */
	private volatile Throwable threadException;

	/** The number of the update cycle currently being executed. */
	private int update;

	/** The array of PageRank scores.*/
	private final double[] score;
	/** The array of new PageRank scores (used during each iteration to write out the new scores).*/
	private final double[] newScore;

	/** The &alpha; parameter (the damping factor). */
	private final double alpha;
	/** The &beta; parameter (used in the weights). */
	private final double beta;

	/** If true a ratio-triangular random walk will be performed. */
	private final boolean ratio;
	
	private final class SimpleUncaughtExceptionHandler implements UncaughtExceptionHandler {
		@Override
		public void uncaughtException( Thread t, Throwable e ) {
			threadException = e;
		}
	}

	/** Creates a new instance of this class.
	 * 
	 * @param graph the graph. 
	 * @param alpha the parameter &alpha; (damping factor).
	 * @param beta the parameter &beta;.
	 * @param maxIter the maximum number of iterations.
	 * @param ratio if true, the ratio-triangular version will be used (otherwise, the mass-triangular one).
	 */
	public LinePageRank( final ImmutableGraph graph, final double alpha, final double beta, final int maxIter, final boolean ratio ) {
		this.graph = graph;
		this.n = graph.numNodes();
		this.m = (int)graph.numArcs();
		simpleUncaughtExceptionHandler = new SimpleUncaughtExceptionHandler();

		this.numberOfThreads = Runtime.getRuntime().availableProcessors();
		this.nextNode = new AtomicInteger();

		this.cumul = new int[ n ];
		this.outdegree = new int[ n ];

		this.alpha = alpha;
		this.beta = beta;
		this.maxIter = maxIter;

		this.score = new double[ m ];
		this.newScore = new double[ m ];
		this.ratio = ratio;
	}

	protected final class IterationThread extends Thread {

		private final ImmutableGraph graph;

		private final ProgressLogger pl;
		
		public double kappa;

		private IterationThread( final ImmutableGraph graph, final ProgressLogger pl ) {
			this.graph = graph;
			this.pl = pl;
			this.kappa = 0;
		}

		public void run() {
			final ImmutableGraph graph = this.graph;
			final int n = LinePageRank.this.n;
			final int[] outdegree = LinePageRank.this.outdegree;
			final int granularity = Math.max( 1024, n >>> 10 );
			final double beta = LinePageRank.this.beta;
			final boolean ratio = LinePageRank.this.ratio;

			int start, end;

			for ( ;; ) {
				final int next = nextNode.getAndAdd( granularity );
				if ( next >= n ) {
					nextNode.getAndAdd( -granularity ); // Put the candle back
					return;
				}
				start = next;
				end = (int)( Math.min( n, (long)start + granularity ) );

				int sum = 0;
				for ( int k = start; k < end; k++ ) {
					final int x = k;
					final int degX = outdegree[ x ];
					int[] succX = graph.successorArray( x );;

					sum += degX;

					for ( int i = 0; i < degX; i++ ) {		
						final int currentArc = cumul[ x ] + i;
						
						final int y = succX[ i ];
						final int degY = outdegree[ y ];
						if( degY < 1 ) {
							kappa += score[ currentArc ];
						}else{
							final int[] succY = graph.successorArray( y );

							int a, b, tr = 0, nonTr = 0;
							////// computing the numbers of triangular
							for ( a = 0, b = 0; a < degX && b < degY; ) {
								if ( succX[ a ] < succY[ b ] )
									a++;
								else if ( succY[ b ] < succX[ a ] ) {
									nonTr++;
									b++;
								}
								else {
									tr++;
									a++;
									b++;
								}
							}
							for ( ; b < degY; b++ )
								nonTr++;

							///// computing weights
							if( tr * nonTr < 1 ) {
								for( b = 0; b < degY; b++ )
									newScore[ cumul[ y ] + b ] += score[ currentArc ] / degY;
							}else{
								double p = 1.0 / ( nonTr * beta + tr);
								double trW = ratio ? p : (1-beta)/tr;
								double nonTrW = ratio ? p*beta : beta / nonTr;
								
								for ( a = 0, b = 0; a < degX && b < degY; ) {
									if ( succX[ a ] < succY[ b ] )
										a++;
									else if ( succY[ b ] < succX[ a ] ) {
										newScore[ cumul[ y ] + b ] += score[ currentArc ] * nonTrW;
										b++;
									}
									else {
										newScore[ cumul[ y ] + b ] += score[ currentArc ] * trW;
										a++;
										b++;
									}
								}
								for ( ; b < degY; b++ ) {
									newScore[ cumul[ y ] + b ] += beta * score[ currentArc ] * nonTrW;
								}
							}
							
							
							
						}
					}
				}
				synchronized ( pl ) {
					pl.update( sum );
				}
			}
		}
	}

	/** Performs an update cycle.
	 * 
	 * @return the L-infty norm of the difference between the new and the old score.
	 */
	private double update() {
		final int m = this.m;
		DoubleArrays.fill( newScore, 0 );

		ProgressLogger pl = new ProgressLogger( LOGGER );
		pl.expectedUpdates = m;
		pl.logInterval = ProgressLogger.TEN_SECONDS;
		pl.itemsName = "arcs";
		pl.start( "Starting update " + ( update + 1 ) + "..." );


		final IterationThread[] thread = new IterationThread[ numberOfThreads ];
		nextNode.set( 0 );
		for ( int i = 0; i < numberOfThreads; i++ ) {
			thread[ i ] = new IterationThread( graph.copy(), pl );
			thread[ i ].setUncaughtExceptionHandler( simpleUncaughtExceptionHandler );
			thread[ i ].start();
		}
		for ( int i = 0; i < numberOfThreads; i++ ) {
			try {
				thread[ i ].join();
			}
			catch ( InterruptedException e ) {
				throw new RuntimeException( e );
			}
		}
		if ( threadException != null )
			throw new RuntimeException( threadException );
		
		double kappa = 0;
		for( int i = 0; i < numberOfThreads; i++ )
			kappa += thread[ i ].kappa;
		
		
		double norm = 0;
		double tmp = 0;
		for( int i = 0; i < m; i++ ) {
			tmp = alpha * ( newScore[ i ] + kappa / m ) + (1-alpha) / m;
			norm = Math.max( norm, Math.abs( tmp - score[ i ] ));
			score[ i ] = tmp;
		}
		pl.done();
		return norm;
	}

	/** Initializes the state before the computation of PageRank.
	 * 
	 */
	public void init() {
		LOGGER.info( "Computing outdegrees..." );
		ProgressLogger pl = new ProgressLogger( LOGGER );
		pl.expectedUpdates = n;
		pl.itemsName = "nodes";
		pl.start();
		int incr = 0;
		for ( NodeIterator iterator = graph.nodeIterator(); iterator.hasNext(); pl.lightUpdate() ) {
			final int node = iterator.nextInt();
			final int deg = iterator.outdegree();
			cumul[ node ] = incr;
			outdegree[ node ] = deg;
			incr += iterator.outdegree();
		}
		pl.done();

		DoubleArrays.fill( score, 1.0 / m );
	}

	public static void main( String[] args ) throws IOException, JSAPException {
		SimpleJSAP jsap = new SimpleJSAP( LinePageRank.class.getName(), "Compute the PageRank of a weighted line graph.", new Parameter[] {
				new FlaggedOption( "alpha", JSAP.DOUBLE_PARSER, "0.85", JSAP.NOT_REQUIRED, 'a', "alpha", "Damping factor" ),
				new FlaggedOption( "beta", JSAP.DOUBLE_PARSER, "0.01", JSAP.NOT_REQUIRED, 'b', "beta", "The beta value." ),
				new FlaggedOption( "maxIter", JSAP.INTEGER_PARSER, "100", JSAP.NOT_REQUIRED, 'm', "maxIter", "The maximum number of iterations." ),
				new Switch( "ratio", 'r', "ratio", "If specified, a ratio-triangular random walk scheme will be used." ),
				new UnflaggedOption( "graph", JSAP.STRING_PARSER, JSAP.REQUIRED, "The basename of loopless graph." ),
				new UnflaggedOption( "dest", JSAP.STRING_PARSER, JSAP.REQUIRED, "The output labelling." ), 
		} );

		JSAPResult jsapResult = jsap.parse( args );
		if ( jsap.messagePrinted() )
			System.exit( 1 );

		LinePageRank lpr = new LinePageRank( BVGraph.load( jsapResult.getString( "graph" ) ), jsapResult.getDouble( "alpha" ), jsapResult.getDouble( "beta" ), jsapResult.getInt( "maxIter" ), jsapResult.userSpecified( "ratio" ));
		lpr.init();
		while( lpr.update() > 1E-6 && lpr.update++ < lpr.maxIter );
		
		BinIO.storeDoubles( lpr.score, jsapResult.getString( "dest" ) );
	}
}