package it.unimi.dsi.law.triangrw;

/*		 
 * Copyright (C) 2011 Paolo Boldi and Marco Rosa 
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 3 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 */

import it.unimi.dsi.Util;
import it.unimi.dsi.fastutil.ints.Int2DoubleMap;
import it.unimi.dsi.fastutil.ints.Int2DoubleOpenHashMap;
import it.unimi.dsi.fastutil.ints.Int2IntMap;
import it.unimi.dsi.fastutil.ints.Int2IntOpenHashMap;
import it.unimi.dsi.fastutil.ints.IntOpenHashSet;
import it.unimi.dsi.fastutil.ints.IntSet;
import it.unimi.dsi.fastutil.io.BinIO;
import it.unimi.dsi.fastutil.longs.Long2ObjectMap;
import it.unimi.dsi.fastutil.longs.Long2ObjectOpenHashMap;
import it.unimi.dsi.fastutil.objects.Object2LongFunction;
import it.unimi.dsi.io.FastBufferedReader;
import it.unimi.dsi.lang.MutableString;
import it.unimi.dsi.logging.ProgressLogger;

import java.io.FileReader;
import java.io.IOException;
import java.io.Serializable;

import org.apache.log4j.Logger;

import com.martiansoftware.jsap.JSAP;
import com.martiansoftware.jsap.JSAPException;
import com.martiansoftware.jsap.JSAPResult;
import com.martiansoftware.jsap.Parameter;
import com.martiansoftware.jsap.SimpleJSAP;
import com.martiansoftware.jsap.Switch;
import com.martiansoftware.jsap.UnflaggedOption;



/** An implementation of {@link ArcSimilarity} assuming that each arc is labelled with text. This class assumes the availability of the following data:
 *  <ul>
 *  	<li>a graph;
 *  	<li>there is a file (hereafter called <em>text file</em>) with one line for each arc (or edge: see below), containing in text form (TAB-separated):
 *  		<ul>
 *  			<li>the number of the source node;
 *  			<li>the number of the target node;
 *  			<li>the text (words are tab-separated, and have been already normalized [downcased, stemmed etc.]).
 *  		</ul>
 *  	<li>there is a file (hereafter called <em>term map</em>) containing a minimal perfect hash function of the terms appearing in the
 *    	text file;
 *    	<li>there is a file (hereafter called <em>frequency file</em>) that contains one integer per term, where the <var>k</var>-th integer
 *    	is the number of arcs that contain that term.
 *  </ul> 
 *  
 *  <p>Let <var>T</var> be the number of terms, <var>f[t,(x,y)]</var> be the number of times the term <var>t</var> appears on the arc <var>(x,y)</var>,
 *  <var>n[t]</var> be the number of arcs that contain the term <var>t</var> and <var>m</var> be the overall number of arcs.
 *  Define, for each term <var>t</var> and each arc <var>(x,y)</var> the weight <var>w[t,(x,y)]</var> as follows:
 *  <ul>
 *  	<li>if <var>f[t,(x,y)]&gt;0</var>, then <var>w[t,(x,y)]=(1+log f[t,(x,y)])*log(m/n[t])</var>;
 *  	<li>otherwise, <var>w[t,(x,y)]=0</var>.
 *  </ul> 
 * 
 *  <p>Let now <var>w[(x,y)]</var> denote the vector of weights for arc <var>(x,y)</var> (a vector of <var>T</var> real numbers).
 *  The similarity between arc <var>(x,y)</var> and arc <var>(x',y')</var> is then defined by <var>w[(x,y)].w[(x',y')]/(a*a')</var>,
 *  where <var>.</var> is the inner vector product and <var>a,a'</var> are the L2-norms of the two vectors.
 *  
 *  <p><b>Symmetric graphs</b>: if the graph is symmetric and texts are also symmetric (i.e., the text on arc <var>(x,y)</var> is the same
 *  as the text on arc <var>(y,x)</var>), the text file only contains the arcs <var>(x,y)</var> with <var>x &lt; y</var>, and a special
 *  flag of symmetricity is set.
 */
public class TextTFIDFArcSimilarity implements Serializable, ArcSimilarity {
	private static final long serialVersionUID = 1L;
	private static final boolean DEBUG = false;
	private static final Logger LOGGER = Util.getLogger( TextTFIDFArcSimilarity.class ); 
	{
		Util.ensureLog4JIsConfigured();
	}
	public Long2ObjectMap<String> arc2text;
	public Object2LongFunction<String> termMap;
	public int[] frequency;
	public boolean symmetric;

	@Override
	public double sim( int from0, int to0, int from1, int to1 ) {
		int temp;
		long arc[] = new long[ 2 ];
		String text;
		Int2IntMap term2freq[] = new Int2IntMap[ 2 ];
		IntSet both = new IntOpenHashSet();
		Int2DoubleMap term2weight[] = new Int2DoubleMap[ 2 ];
		double norm[] = new double[ 2 ];
		double innerProduct = 0;
		
		if ( symmetric && from0 > to0 ) { temp = from0; from0 = to0; to0 = temp; }
		if ( symmetric && from1 > to1 ) { temp = from1; from1 = to1; to1 = temp; }
		arc[ 0 ] = ( (long)from0 << 32 ) | to0;
		arc[ 1 ] = ( (long)from1 << 32 ) | to1;
		for ( int i = 0; i < 2; i++ ) {
			text = arc2text.get( arc[ i ] );
			if ( DEBUG ) System.out.println( text );
			if ( text == null ) return 0;
			term2freq[ i ] = new Int2IntOpenHashMap();
			term2freq[ i ].defaultReturnValue( 0 );
			// Tokenize, compute frequencies and common set
			int from = 0, to;
			while ( ( to = text.indexOf( ' ', from ) ) >= 0 ) {
				String term = text.substring( from, to );
				int termNumber = (int)termMap.getLong( term );
				if ( DEBUG ) System.out.println( term + " -> " + termNumber + " (freq: " + frequency[ termNumber ] + ")" );
				if ( i > 0 && !both.contains( termNumber ) && term2freq[ 0 ].containsKey( termNumber ) ) both.add( termNumber );
				term2freq[ i ].put( termNumber, term2freq[ i ].get( termNumber ) + 1 );
				from = to + 1;
			}
			if ( DEBUG ) System.out.println( "Frequencies: " + term2freq[ i ] );
		}
		if ( DEBUG ) System.out.println( "Both: " + both );
		// Compute weights and norms
		for ( int i = 0; i < 2; i++ ) {
			term2weight[ i ] = new Int2DoubleOpenHashMap();
			for ( int termNumber: term2freq[ i ].keySet() ) {
				double weight = ( 1 + Math.log( term2freq[ i ].get( termNumber ) ) ) * Math.log( (double)arc2text.size() / frequency[ termNumber ] );
				norm[ i ] += weight * weight;
				if ( both.contains( termNumber ) ) term2weight[ i ].put( termNumber, weight );
			}
		}
		if ( DEBUG ) System.out.println( "Norm first vector: " + Math.sqrt( norm[ 0 ] ) );
		if ( DEBUG ) System.out.println( "Norm second vector: " + Math.sqrt( norm[ 1 ] ) );
		if ( DEBUG ) System.out.println( "Term2weight first vector: " + term2weight[ 0 ] );
		if ( DEBUG ) System.out.println( "Term2weight second vector: " + term2weight[ 1 ] );
		// Compute inner product
		for ( int termNumber: both ) innerProduct += term2weight[ 0 ].get( termNumber ) * term2weight[ 1 ].get( termNumber );
		if ( DEBUG ) System.out.println( "Inner product: " + innerProduct );
		if ( norm[ 0 ] <= 0 || norm[ 1 ] <= 0 ) return 0; // empty documents?
		// Return result
		return innerProduct / Math.sqrt( norm[ 0 ] * norm[ 1 ] );
	}
	
	@SuppressWarnings( "unchecked" )
	public static void main( String arg[] ) throws SecurityException, JSAPException, NoSuchMethodException, IOException, ClassNotFoundException {
		final SimpleJSAP jsap = new SimpleJSAP( TextTFIDFArcSimilarity.class.getName(), "Creates a TextTFIDFSimilarity object.",
				new Parameter[] {
					new Switch( "symmetric", 's', "symmetric", "The graph is symmetric and only the pairs (x,y) with x<y are listed in the text file." ),
					new UnflaggedOption( "arcTextFile", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.REQUIRED, JSAP.NOT_GREEDY, "The text file containing one line per arc (or edge); format SOURCE [TAB] TARGET [TAB] TEXT." ),
					new UnflaggedOption( "termMapFile", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.REQUIRED, JSAP.NOT_GREEDY, "The minimal perfect term map." ),
					new UnflaggedOption( "termFrequencyFile", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.REQUIRED, JSAP.NOT_GREEDY, "The frequency file (containing one integer per term, the number of arcs (or edges) containing it)." ),
					new UnflaggedOption( "resultFile", JSAP.STRING_PARSER, JSAP.NO_DEFAULT, JSAP.REQUIRED, JSAP.NOT_GREEDY, "The name of the file where the TextTFIDF object will be saved." )
		});

		JSAPResult jsapResult = jsap.parse( arg );
		if ( jsap.messagePrinted() ) return;
		
		final String arcTextFile = jsapResult.getString( "arcTextFile" );
		final String termMapFile = jsapResult.getString( "termMapFile" );
		final String termFrequencyFile = jsapResult.getString( "termFrequencyFile" );
		final boolean symmetric = jsapResult.getBoolean( "symmetric" );
		final String resultFile = jsapResult.getString( "resultFile" );

		ProgressLogger pl = new ProgressLogger();
		
		TextTFIDFArcSimilarity result = new TextTFIDFArcSimilarity();
		
		result.symmetric = symmetric;
		LOGGER.info(  "Loading term map" );
		result.termMap = (Object2LongFunction<String>)BinIO.loadObject( termMapFile );
		LOGGER.info(  "Loading term frequencies" );
		result.frequency = BinIO.loadInts( termFrequencyFile );
		result.arc2text = new Long2ObjectOpenHashMap<String>();
		
		pl.itemsName = "pairs";
		pl.start( "Reading arc text" );
		FastBufferedReader reader = new FastBufferedReader( new FileReader( arcTextFile ) );
		MutableString line = new MutableString();
		int firstTab, secondTab;
		
		while ( reader.readLine( line ) != null ) {
			if ( ( firstTab = line.indexOf( '\t' ) ) < 0 || ( secondTab = line.indexOf( '\t', firstTab + 1 ) ) < 0 ) {
				System.err.println( "Malformed line: <" + line + ">" );
				return;
			}
			int from = Integer.parseInt( line.substring( 0, firstTab ).toString() );
			int to = Integer.parseInt( line.substring( firstTab + 1, secondTab ).toString() );
			result.arc2text.put( ( (long)from << 32 ) | to, line.substring( secondTab + 1 ).toString() );
			pl.lightUpdate();
		}
		pl.done();
		LOGGER.info(  "Saving result" );
		BinIO.storeObject( result, resultFile );
	}

}
